@@ -1,12 +1,19 @@
 ;;; Foreign function interface for CLISP
 ;;; Bruno Haible 19.2.1995
+;;; Sam Steingold 1998-2002
+
+#+UNICODE
+(progn
+  (in-package "EXT")
+  (export '(custom::*foreign-encoding*) "CUSTOM")
+  (export '(custom::*foreign-encoding*) "EXT"))
 
 (use-package '("COMMON-LISP" "EXT") "FFI")
 (in-package "FFI")
 
 (export '(def-c-type def-c-var
           def-c-call-out def-call-out #+AMIGA def-lib-call-out
-          def-c-call-in def-call-in
+          def-c-call-in def-call-in default-foreign-language
           c-lines
           nil boolean character char uchar short ushort int uint long ulong
           uint8 sint8 uint16 sint16 uint32 sint32 uint64 sint64
@@ -14,9 +21,9 @@
           c-pointer c-string c-struct c-union c-array c-array-max
           c-function c-ptr c-ptr-null c-array-ptr
           def-c-enum def-c-struct element deref slot cast typeof
-          sizeof bitsizeof
-          validp
-          #+UNICODE *foreign-encoding*
+          sizeof bitsizeof c-var-address offset
+          validp foreign-address-null
+          with-foreign-object with-c-var with-foreign-string
           foreign-address foreign-variable foreign-function))
 
 (eval-when (load compile eval)
@@ -24,6 +31,8 @@
   (import (intern "*COUTPUT-STREAM*" "COMPILER"))
   (import (intern "*FFI-MODULE*" "COMPILER"))
   (import (intern "FINALIZE-COUTPUT-FILE" "COMPILER"))
+  (import (intern "TEXT" "SYSTEM")) ; messages
+  (import (intern "CHECK-SYMBOL" "SYSTEM")) ; error checking
   (import (intern "DEPARSE-C-TYPE" "SYSTEM")) ; called by DESCRIBE
   (import (intern "FOREIGN-FUNCTION-IN-ARG-COUNT" "SYSTEM")) ; called by SYS::FUNCTION-SIGNATURE
 )
@@ -80,7 +89,7 @@
     (write-char #\" s)
     (map nil #'(lambda (c)
                  (cond ((eql c #\Null)
-                        (error (ENGLISH "Cannot map string ~S to C since it contains a character ~S")
+                        (error (TEXT "Cannot map string ~S to C since it contains a character ~S")
                                string c))
                        ((eq c #\Newline)
                         (write-char #\\ s) (write-char #\n s))
@@ -118,7 +127,7 @@
               (unless (and (consp subspec)
                            (eql (length subspec) 2)
                            (symbolp (first subspec)))
-                (error (ENGLISH "Invalid ~S component: ~S")
+                (error (TEXT "Invalid ~S component: ~S")
                        form-type subspec))
               (parse-c-type (second subspec)))
             spec-list)))
@@ -169,14 +178,14 @@
     (if (symbolp typespec)
       (multiple-value-bind (c-type found) (gethash typespec *c-type-table*)
         (unless found
-          (error (ENGLISH "Incomplete FFI type ~S is not allowed here.")
+          (error (TEXT "Incomplete FFI type ~S is not allowed here.")
                  typespec))
         (when name (setf (gethash name *c-type-table*) c-type))
         c-type)
-      (error (ENGLISH "FFI type should be a symbol, not ~S")
+      (error (TEXT "FFI type should be a symbol, not ~S")
              typespec))
     (flet ((invalid (typespec)
-             (error (ENGLISH "Invalid FFI type: ~S")
+             (error (TEXT "Invalid FFI type: ~S")
                     typespec))
            (dimp (dim) (typep dim '(integer 0 *))))
       (case (first typespec)
@@ -239,10 +248,10 @@
   (let ((alist '()))
     (dolist (option options)
       (unless (and (consp option) (member (first option) keywords))
-        (error (ENGLISH "Invalid option in ~S: ~S")
+        (error (TEXT "Invalid option in ~S: ~S")
                whole option))
       (when (assoc (first option) alist)
-        (error (ENGLISH "Only one ~S option is allowed: ~S")
+        (error (TEXT "Only one ~S option is allowed: ~S")
               (first option) whole))
       (push option alist))
     alist))
@@ -268,6 +277,13 @@
     (:STDC ff-language-ansi-c)
     (:STDC-STDCALL (+ ff-language-ansi-c ff-language-stdcall))))
 
+;; the default foreign language
+(defvar *foreign-language* nil)
+
+(defmacro default-foreign-language (lang)
+  (language-to-flag lang)       ; error checking
+  `(eval-when (load compile eval) (setq *foreign-language* ',lang)))
+
 ;; get the even (start=0) or odd (start=1) elements of the simple vector
 (defun split-c-fun-arglist (args start)
   (do ((ii start (+ ii 2)) (res '()))
@@ -282,7 +298,7 @@
                       (unless (and (listp argspec)
                                    (symbolp (first argspec))
                                    (<= 2 (length argspec) #-AMIGA 4 #+AMIGA 5))
-                        (error (ENGLISH "Invalid parameter specification in ~S: ~S")
+                        (error (TEXT "Invalid parameter specification in ~S: ~S")
                                whole argspec))
                       (let* ((argtype (parse-c-type (second argspec)))
                              (argmode (if (cddr argspec) (third argspec) ':IN))
@@ -293,6 +309,11 @@
                                                  (eq argmode ':OUT))
                                              ':ALLOCA
                                              ':NONE))))
+                        ;; see FOREIGN-CALL-OUT in foreign.d
+                        (when (and (or (eq argmode :OUT) (eq argmode :IN-OUT))
+                                   (not (eq (ctype-type argtype) 'C-PTR)))
+                          (warn (TEXT "~s argument ~s is not a pointer in ~s")
+                                argmode argtype whole))
                         (list argtype
                               (+ (ecase argmode
                                    ((:IN :READ-ONLY) 0)
@@ -317,25 +338,24 @@
        (let ((languages (assoc ':language alist)))
          (if languages
            (reduce #'+ (rest languages) :key #'language-to-flag)
-           ff-language-c)))))   ; Default is K&R C
+           (language-to-flag
+            (or *foreign-language*
+                (progn
+                  (warn (TEXT "~s: No ~s argument and no ~s form in this compilation unit; ~s assumed now and for the rest of this unit")
+                        whole :language 'default-foreign-language :stdc)
+                  (setq *foreign-language* :STDC))))))))) ; Default is ANSI C
 
 (defun parse-foreign-name (name)
   (unless (stringp name)
-    (error (ENGLISH "The name must be a string, not ~S")
+    (error (TEXT "The name must be a string, not ~S")
            name))
   (if (c-ident-p name)
     name
-    (error (ENGLISH "The name ~S is not a valid C identifier")
+    (error (TEXT "The name ~S is not a valid C identifier")
            name)))
 
-(defun check-symbol (whole &optional (name (second whole)))
-  (unless (symbolp name)
-    (sys::error-of-type 'sys::source-program-error
-      (ENGLISH "~S: this is not a symbol: ~S")
-      (first whole) name)))
-
 (defmacro DEF-C-TYPE (&whole whole name typespec)
-  (check-symbol whole)
+  (check-symbol (first whole) name)
   `(EVAL-WHEN (LOAD COMPILE EVAL)
      (PARSE-C-TYPE ',typespec ',name)
      ',name))
@@ -507,7 +527,8 @@
     (uint64 (format nil "uint64 ~A" name))
     (single-float (format nil "float ~A" name))
     (double-float (format nil "double ~A" name))
-    ((c-pointer c-string) (format nil "char* ~A" name))
+    (c-string (format nil "char* ~A" name))
+    (c-pointer (format nil "void* ~A" name))
     (t (if (gethash c-type *type-table*)
          (format nil "~A ~A" (gethash c-type *type-table*) name)
          (macrolet ((with-to-c ((class typename) &body body)
@@ -550,7 +571,7 @@
                                                 c-t (gensym "arg")))
                                              (split-c-fun-arglist
                                               (svref c-type 2) 0)))))
-             (t (error (ENGLISH "illegal foreign data type ~S")
+             (t (error (TEXT "illegal foreign data type ~S")
                        c-type))))))))
 
 (defun prepare-module ()
@@ -564,18 +585,23 @@
 (defun finalize-coutput-file ()
   (when *ffi-module*
     (format *coutput-stream* "~%")
-    (format *coutput-stream* "subr_ module__~A__subr_tab[1];~%" *c-name*)
+    (format *coutput-stream* "subr_t module__~A__subr_tab[1];~%" *c-name*)
     (format *coutput-stream* "uintC module__~A__subr_tab_size = 0;~%" *c-name*)
-    (format *coutput-stream* "subr_initdata module__~A__subr_tab_initdata[1];~%" *c-name*)
+    (format *coutput-stream* "subr_initdata_t module__~A__subr_tab_initdata[1];~%"
+            *c-name*)
     (format *coutput-stream* "~%")
     (let ((count (hash-table-count *object-table*)))
       (if (zerop count)
         (progn
           (format *coutput-stream* "object module__~A__object_tab[1];~%" *c-name*)
-          (format *coutput-stream* "object_initdata module__~A__object_tab_initdata[1];~%" *c-name*))
+          (format *coutput-stream*
+                  "object_initdata_t module__~A__object_tab_initdata[1];~%" *c-name*))
         (let ((v (make-array count)))
-          (format *coutput-stream* "object module__~A__object_tab[~D];~%" *c-name* count)
-          (format *coutput-stream* "object_initdata module__~A__object_tab_initdata[~D] = {~%" *c-name* count)
+          (format *coutput-stream* "object module__~A__object_tab[~D];~%"
+                  *c-name* count)
+          (format *coutput-stream*
+                  "object_initdata_t module__~A__object_tab_initdata[~D] = {~%"
+                  *c-name* count)
           (dohash (key value *object-table*)
             (declare (ignore key))
             (setf (svref v (cdr value)) (car value)))
@@ -603,10 +629,10 @@
               (to-c-typedecl (svref (second function) 1)
                              (format nil "(~A)()" (first function)))))
     (format *coutput-stream* "~%void module__~A__init_function_1(module)
-  var module_* module;~%{ }~%"
+  var module_t* module;~%{ }~%"
             *c-name*)
     (format *coutput-stream* "~%void module__~A__init_function_2(module)
-  var module_* module;~%{~%"
+  var module_t* module;~%{~%"
             *c-name*)
     (dolist (variable *variable-list*)
       (format *coutput-stream*
@@ -657,12 +683,12 @@
     (to-c-name lisp-name)))
 
 (defmacro DEF-C-VAR (&whole whole name &rest options)
-  (check-symbol whole)
+  (check-symbol (first whole) name)
   (let* ((alist (parse-options options '(:name :type :read-only :alloc) whole))
          (c-name (foreign-name name (assoc ':name alist)))
          (type (second (or (assoc ':type alist)
                            (sys::error-of-type 'sys::source-program-error
-                                  (ENGLISH "~S: ~S option missing in ~S")
+                                  (TEXT "~S: ~S option missing in ~S")
                                   'def-c-var ':type whole))))
          (read-only (second (assoc ':read-only alist)))
          (flags (+ (if read-only fv-flag-readonly 0)
@@ -678,7 +704,7 @@
          |#
         )
     `(PROGN
-       (EVAL-WHEN (COMPILER::COMPILE-ONCE-ONLY) (NOTE-C-VAR ',c-name ',type ',flags))
+       (EVAL-WHEN (COMPILE) (NOTE-C-VAR ',c-name ',type ',flags))
        #|
        (LET ((FVAR (FFI::LOOKUP-FOREIGN-VARIABLE ',c-name (PARSE-C-TYPE ',type))))
          (DEFUN ,getter-function-name () (FFI::FOREIGN-VALUE FVAR))
@@ -703,23 +729,56 @@
 
 (defsetf ffi::foreign-value ffi::set-foreign-value)
 
+(defun foreign-address-null (fadr)
+  (zerop (foreign-address-unsigned fadr)))
+
+;; ============================ Stack allocation ============================
+
+;; Allocate arbitrarily complex structure on the stack,
+;; but allocate only (sizeof c-type) bytes when called without initarg!
+;; C-ARRAY-MAX, "" and #() are thus your friends for creation of empty arrays.
+;; (with-c-var (v '(c-ptr  (c-array     uint8 32)))     (cast v 'c-pointer))
+;;                -> null-pointer, 4 bytes
+;; (with-c-var (v '(c-ptr  (c-array-max uint8 32)) #()) (cast v 'c-pointer))
+;;                -> non-null,  4+32 bytes
+;; (with-c-var (v '(c-ptr-null (c-array uint8 32)) nil) (cast v 'c-pointer))
+;;                -> null-pointer, 4 bytes
+
+;; c-type is evaluated. This is particularly useful for variable sized arrays
+;; using: `(c-array uint8 ,(length foo))
+(defmacro with-foreign-object ((var c-type &optional (init nil init-p))
+                               &body body)
+  `(EXEC-ON-STACK (LAMBDA (,var) ,@body)
+                  (PARSE-C-TYPE ,c-type)
+                  . ,(if init-p (list init))))
+
+;; symbol-macro based interface (like DEF-C-VAR)
+(defmacro with-c-var ((var c-type &optional (init nil init-p)) &body body)
+  (let ((fv (gensym (symbol-name var))))
+    `(EXEC-ON-STACK
+      (LAMBDA (,fv) (SYMBOL-MACROLET ((,var (FFI::FOREIGN-VALUE ,fv))) ,@body))
+      (PARSE-C-TYPE ,c-type)
+      . ,(if init-p (list init)))))
+
 ;; ============================ named C functions ============================
 
 (defmacro DEF-C-CALL-OUT (name &rest options)
+  (warn (TEXT "~s is deprecated, use ~s instead")
+        'def-c-call-out 'def-call-out)
   `(DEF-CALL-OUT ,name ,@options (:LANGUAGE :STDC)))
 
 (defmacro DEF-CALL-OUT (&whole whole name &rest options)
-  (check-symbol whole)
+  (check-symbol (first whole) name)
   (let* ((alist (parse-options options '(:name :arguments :return-type :language) whole))
          (parsed-function (parse-c-function alist whole))
          (signature (argvector-to-signature (svref parsed-function 2)))
          (c-name (foreign-name name (assoc ':name alist))))
     (setq alist (remove (assoc ':name alist) alist))
     `(PROGN
-       (EVAL-WHEN (COMPILER::COMPILE-ONCE-ONLY) (NOTE-C-FUN ',c-name ',alist ',whole))
+       (EVAL-WHEN (COMPILE) (NOTE-C-FUN ',c-name ',alist ',whole))
        (LET ()
          (SYSTEM::REMOVE-OLD-DEFINITIONS ',name)
-         (EVAL-WHEN (COMPILE) (COMPILER::C-DEFUN ',name ',signature))
+         (COMPILER::EVAL-WHEN-COMPILE (COMPILER::C-DEFUN ',name ',signature))
          (SYSTEM::%PUTD ',name
            (FFI::LOOKUP-FOREIGN-FUNCTION ',c-name
                                          (PARSE-C-FUNCTION ',alist ',whole))))
@@ -732,7 +791,7 @@
 
 #+AMIGA
 (defmacro DEF-LIB-CALL-OUT (&whole whole name library &rest options)
-  (check-symbol whole)
+  (check-symbol (first whole) name)
   (let* ((alist (parse-options options '(:name :offset :arguments :return-type) whole))
          (parsed-function
           (parse-c-function (remove (assoc ':name alist) alist) whole))
@@ -741,7 +800,7 @@
          (offset (second (assoc ':offset alist))))
     `(LET ()
        (SYSTEM::REMOVE-OLD-DEFINITIONS ',name)
-       (EVAL-WHEN (COMPILE) (COMPILER::C-DEFUN ',name ',signature))
+       (COMPILER::EVAL-WHEN-COMPILE (COMPILER::C-DEFUN ',name ',signature))
        (SYSTEM::%PUTD ',name
          (FFI::FOREIGN-LIBRARY-FUNCTION ',c-name
            (FFI::FOREIGN-LIBRARY ',library)
@@ -750,16 +809,18 @@
        ',name)))
 
 (defmacro DEF-C-CALL-IN (name &rest options)
+  (warn (TEXT "~s is deprecated, use ~s instead")
+        'def-c-call-in 'def-call-in)
   `(DEF-CALL-IN ,name ,@options (:LANGUAGE :STDC)))
 
 (defmacro DEF-CALL-IN (&whole whole name &rest options)
-  (check-symbol whole)
+  (check-symbol (first whole) name)
   (let* ((alist (parse-options
                  options '(:name :arguments :return-type :language) whole))
          (c-name (foreign-name name (assoc ':name alist))))
     (setq alist (remove (assoc ':name alist) alist))
     `(PROGN
-       (EVAL-WHEN (COMPILER::COMPILE-ONCE-ONLY)
+       (EVAL-WHEN (COMPILE)
          (NOTE-C-CALL-IN ',name ',c-name ',alist ',whole))
        ',name)))
 
@@ -832,7 +893,7 @@
           (mapc #'(lambda (argtype argflag argname)
                     (when (flag-set-p argflag flag-output)
                       (unless (eq (ctype-type argtype) 'C-PTR)
-                        (error (ENGLISH "~S: :OUT argument is not a pointer: ~S")
+                        (error (TEXT "~S: :OUT argument is not a pointer: ~S")
                                'DEF-CALL-IN argtype))
                       (format *coutput-stream* "  ~A~A(~A,~A,~A);~%"
                               (if (eql outargcount 0) ""
@@ -873,7 +934,7 @@
   (count-inarguments (sys::%record-ref obj 3)))
 
 (defmacro def-c-enum (&whole whole name &rest items)
-  (check-symbol whole)
+  (check-symbol (first whole) name)
   (let ((forms '())
         (next-value 0))
     (dolist (item items)
@@ -894,9 +955,11 @@
 ;; (element (foreign-value x) ...) --> (foreign-value (%element x ...))
 ;; (deref (foreign-value x))       --> (foreign-value (%deref x))
 ;; (slot (foreign-value x) ...)    --> (foreign-value (%slot x ...))
+;; (cast (foreign-value x) ...)    --> (foreign-value (%cast x ...))
+;; (offset (foreign-value x) ...)  --> (foreign-value (%offset x ...))
 (flet ((err (whole)
          (sys::error-of-type 'sys::source-program-error
-           (ENGLISH "~S is only allowed after ~S: ~S")
+           (TEXT "~S is only allowed after ~S: ~S")
            (first whole) 'FOREIGN-VALUE whole))
        (foreign-place-p (place type)
          (and (consp place) (eq (first place) type) (eql (length place) 2))))
@@ -920,6 +983,18 @@
     (if (foreign-place-p place 'FOREIGN-VALUE)
       `(FOREIGN-VALUE (%CAST ,(second place) (PARSE-C-TYPE ,type)))
       (err `(cast ,place ,type))))
+  (defmacro offset (place offset type &environment env)
+    (setq place (macroexpand place env))
+    (if (foreign-place-p place 'FOREIGN-VALUE)
+      `(FOREIGN-VALUE (%OFFSET ,(second place) ,offset (PARSE-C-TYPE ,type)))
+      (err `(offset ,place ,offset ,type))))
+  ;; The equivalent of (FOREIGN-ADDRESS fvar) for c-places:
+  ;; (c-var-address (foreign-value x)) --> (foreign-address x)
+  (defmacro c-var-address (place &environment env)
+    (setq place (macroexpand place env))
+    (if (foreign-place-p place 'FOREIGN-VALUE)
+      `(FOREIGN-ADDRESS ,(second place))
+      (err `(c-var-address ,place))))
   ;; Similarly for TYPEOF.
   ;; (typeof (foreign-value x)) --> (deparse-c-type (foreign-type x))
   (defmacro typeof (place &environment env)
@@ -959,3 +1034,10 @@
   (defsetf system::foreign-encoding system::set-foreign-encoding))
 
 ;; ===========================================================================
+
+;;; for self-test
+;;; commented out since CLISP does not need them;
+;;; they are defined by the test suite.
+;; (def-call-out ffi_identity (:arguments (obj int))
+;;   (:return-type int) (:language :stdc))
+;; (def-c-var ffi_user_pointer (:type c-pointer))
